#ifndef PSPA_SEEN
#define  PSPA_SEEN

#include <cstdlib>
#include <list>

#include <Wt/WApplication>
#include <Wt/WEnvironment>
#include <Wt/WContainerWidget>
#include <Wt/WText>
#include <Wt/WPanel>
#include <Wt/WComboBox>
#include <Wt/WPushButton>
#include <Wt/WDialog>
#include <Wt/WPdfImage>
#include <Wt/Chart/WCartesianChart>
#include <Wt/WLength>
#include <Wt/WButtonGroup>

#include "GWt_dialog.h"
#include "dataManager.h"
#include "GWt_console.h"
#include "GWt_abstractElementFactory.h"
#include "GWt_softwarePanel.h"
#include "GWt_ligneFaisceau.h"

using namespace Wt;

class PspaApplication : public WApplication
{
        
 private :
    
  dataManager* dtmanage_;
  string nameOfCase_;
  string workingDir_;
  WDialog* dialogSave_;
  WDialog* dialogOpen_;
  WLineEdit* saveNameEdit_;
  WLineEdit* openNameEdit_;
  GWt_LigneFaisceau* beamLine_;
    
  WWidget* globalParam_;
  GWt_dialog* graphicsDialog_;
  WButtonGroup *group_;
  vector<GWt_dialog*> eDialog_;
  
  WText* probleme_;
  WComboBox* choixElementDessin_;
  WComboBox* choixTypeDessinFaisceau_;
  WComboBox* choixAbsPhase_;
  WComboBox* choixOrdPhase_;
  WComboBox* choixEnveloppeDessin_;
  WComboBox* choixHistoDessin_;

  WContainerWidget* toto_;
  GWt_softwarePanel* executeWidget_;
  WFileUpload* uploadFileSelectorWidget_;
  GWt_console* console_;
  GWt_abstractElementFactory* abstractElementFactory_;
      
  WText* createTitle(const WString&);
  WWidget* createPalette();
  void createBeamLine();
  WWidget* createGlobalParamWidget();
  WWidget* createExecuteWidget();
  WContainerWidget* createDrawingWidget();
 
  void closeGraphicDialog();
  
  void createDragImage(const char *url,const char *smallurl,const char *mimeType,WContainerWidget *p,WString());
  void sauver();
  void dialogSaveDone(WDialog::DialogCode code);
  void restaurer();
  void chargerConfig();
  void openFileSelector();
  void fileTooLarge();
  
  void dialogOnGraphics();
  void setModelComponentForPhaseSpace(WStandardItemModel* model, vector<bareParticle>& partic, unsigned indexParticleComponent, unsigned modelComponent); 
void faireDessinParmela(WContainerWidget* toto, particleBeam* beam, unsigned indexAbs, unsigned indexOrd, string namex, string namey  );
  void faireDessinTransport(WContainerWidget* toto, particleBeam* beam, unsigned indexAbs, unsigned indexOrd, string namex, string namey);
  void faireDessinEnveloppe(WContainerWidget* toto, string type);
  void scatterPlot1D(WContainerWidget* toto, vector<double>& xcor, vector<double>& ycor);
    
  bool removePathFromConfigName(string& config);
  bool removeExtensionFromConfigName(string& config);
  
  void dessiner();
  void dessinerEnveloppe();
  void dessinerPhaseSpace();
  void dessinerHistogramme();

 public :
  
  PspaApplication(const WEnvironment& env);
  
  string getSelection();
  void faireDessin();

  inline dataManager* getDataManager() const  {
    return dtmanage_;
  }
  
  inline WWidget* getGlobalParam() {
    return globalParam_;
  }

  inline GWt_softwarePanel* getExecuteWidget() {
    return executeWidget_;
  }

  inline GWt_LigneFaisceau* getBeamLine() {
    return beamLine_;
  }

  inline GWt_abstractElementFactory* getAbstractElementFactory() {
    return abstractElementFactory_;
  }

    inline string getWorkingDir() const {
        return workingDir_;
    }

};
#endif
